@@ -704,6 +704,7 @@ int parse_new_id3(mpg123_handle *fr, unsigned long first4bytes)
 		,1) \
 	)
 	/* id3v2.3 does not store synchsafe frame sizes, but synchsafe tag size - doh! */
+	/* Remember: bytes_to_long() can yield ULONG_MAX on 32 bit platforms! */
 	#define bytes_to_long(buf,res) \
 	( \
 		major == 3 ? \
@@ -772,16 +773,25 @@ int parse_new_id3(mpg123_handle *fr, unsigned long first4bytes)
 			if((ret2 = fr->rd->read_frame_body(fr,tagdata,length)) > 0)
 			{
 				unsigned long tagpos = 0;
+				/* bytes of frame title and of framesize value */
+				unsigned int head_part = fr->id3v2.version > 2 ? 4 : 3;
+				unsigned int flag_part = fr->id3v2.version > 2 ? 2 : 0;
+				/* The amount of bytes that are unconditionally read for each frame: */
+				/* ID, size, flags. */
+				unsigned int framebegin = head_part+head_part+flag_part;
 				debug1("ID3v2: have read at all %lu bytes for the tag now", (unsigned long)length+6);
 				/* going to apply strlen for strings inside frames, make sure that it doesn't overflow! */
 				tagdata[length] = 0;
 				if(flags & EXTHEAD_FLAG)
 				{
 					debug("ID3v2: skipping extended header");
-					if(!bytes_to_long(tagdata, tagpos))
+					if(!bytes_to_long(tagdata, tagpos) || tagpos >= length)
 					{
 						ret = 0;
-						if(NOQUIET) error4("Bad (non-synchsafe) tag offset: 0x%02x%02x%02x%02x", tagdata[0], tagdata[1], tagdata[2], tagdata[3]);
+						if(NOQUIET)
+							error4( "Bad (non-synchsafe/too large) tag offset:"
+								"0x%02x%02x%02x%02x"
+							,	tagdata[0], tagdata[1], tagdata[2], tagdata[3] );
 					}
 				}
 				if(ret > 0)
@@ -789,13 +799,12 @@ int parse_new_id3(mpg123_handle *fr, unsigned long first4bytes)
 					char id[5];
 					unsigned long framesize;
 					unsigned long fflags; /* need 16 bits, actually */
-					/* bytes of frame title and of framesize value */
-					int head_part = fr->id3v2.version > 2 ? 4 : 3;
-					int flag_part = fr->id3v2.version > 2 ? 2 : 0;
 					id[4] = 0;
-					/* pos now advanced after ext head, now a frame has to follow */
+					/* Pos now advanced after ext head, now a frame has to follow. */
+					/* Note: tagpos <= length, which is 28 bit integer, so both */
+					/* far away from overflow for adding known small values. */
 					/* I want to read at least one full header now. */
-					while(tagpos <= length-head_part-head_part-flag_part)
+					while(length >= tagpos+framebegin)
 					{
 						int i = 0;
 						unsigned long pos = tagpos;
@@ -828,12 +837,7 @@ int parse_new_id3(mpg123_handle *fr, unsigned long first4bytes)
 								break;
 							}
 							if(VERBOSE3) fprintf(stderr, "Note: ID3v2 %s frame of size %lu\n", id, framesize);
-							tagpos += head_part + framesize; /* the important advancement in whole tag */
-							if(tagpos > length-flag_part)
-							{
-								if(NOQUIET) error("Whoa! ID3v2 frame claims to be larger than the whole rest of the tag.");
-								break;
-							}
+							tagpos += head_part;
 							pos += head_part;
 							if(fr->id3v2.version > 2)
 							{
@@ -842,6 +846,13 @@ int parse_new_id3(mpg123_handle *fr, unsigned long first4bytes)
 								tagpos += 2;
 							}
 							else fflags = 0;
+
+							if(length - tagpos < framesize)
+							{
+								if(NOQUIET) error("Whoa! ID3v2 frame claims to be larger than the whole rest of the tag.");
+								break;
+							}
+							tagpos += framesize; /* the important advancement in whole tag */
 							/* for sanity, after full parsing tagpos should be == pos */
 							/* debug4("ID3v2: found %s frame, size %lu (as bytes: 0x%08lx), flags 0x%016lx", id, framesize, framesize, fflags); */
 							/* %0abc0000 %0h00kmnp */
@@ -881,7 +892,7 @@ int parse_new_id3(mpg123_handle *fr, unsigned long first4bytes)
 									/* de-unsync: FF00 -> FF; real FF00 is simply represented as FF0000 ... */
 									/* damn, that means I have to delete bytes from withing the data block... thus need temporal storage */
 									/* standard mandates that de-unsync should always be safe if flag is set */
-									realdata = (unsigned char*) malloc(framesize); /* will need <= bytes */
+									realdata = (unsigned char*) malloc(framesize+1); /* will need <= bytes, plus a safety zero */
 									if(realdata == NULL)
 									{
 										if(NOQUIET) error("ID3v2: unable to allocate working buffer for de-unsync");
@@ -898,6 +909,8 @@ int parse_new_id3(mpg123_handle *fr, unsigned long first4bytes)
 										}
 									}
 									realsize = opos;
+									/* Append a zero to keep strlen() safe. */
+									realdata[realsize] = 0;
 									debug2("ID3v2: de-unsync made %lu out of %lu bytes", realsize, framesize);
 								}
 								pos = 0; /* now at the beginning again... */
@@ -925,14 +938,22 @@ int parse_new_id3(mpg123_handle *fr, unsigned long first4bytes)
 										if(fr->rva.level[rva_mode] <= rva2+1)
 										{
 											pos += strlen((char*) realdata) + 1;
-											if(realdata[pos] == 1)
+											if(pos >= realsize)
+											{
+												if(NOQUIET)
+												error("bad RVA2 tag (non-terminated identification)");
+											}
+											else if(realdata[pos] == 1)
 											{
 												++pos;
 												/* only handle master channel */
 												debug("ID3v2: it is for the master channel");
 												/* two bytes adjustment, one byte for bits representing peak - n bytes, eh bits, for peak */
-												/* 16 bit signed integer = dB * 512  ... the double cast is needed to preserve the sign of negative values! */
-												fr->rva.gain[rva_mode] = (float) ( (((short)((signed char)realdata[pos])) << 8) | realdata[pos+1] ) / 512;
+												/* 16 bit signed integer = dB * 512. Do not shift signed integers! Multiply instead.
+												   Also no implementation-defined casting. Reinterpret the pointer to signed char, then do
+												   proper casting. */
+												fr->rva.gain[rva_mode] = (float) (
+													((short)((signed char*)realdata)[pos]) * 256 + (short)realdata[pos+1] ) / 512;
 												pos += 2;
 												if(VERBOSE3) fprintf(stderr, "Note: RVA value %fdB\n", fr->rva.gain[rva_mode]);
 												/* heh, the peak value is represented by a number of bits - but in what manner? Skipping that part */
